import os.path
import numpy as np
import pandas as pd
import pickle
import random
import astropy.units as u
from datetime import datetime
from sklearn.neighbors import KDTree

from mathMB import fit_model
from MESSENGERuvvs import MESSENGERdata

from .ModelResults import ModelResult
from .database_connect import database_connect
from .Input import Input
from .Output import Output


def determine_model_type(data):
    # Determins whether model is one complete orbit or a more complicated query
    if len(data.data.orbit.unique() == 1):
        orb = data.data.orbit.unqiue()[0]
        mdata = MESSENGERdata(data.species, f'orbit = {orb}')
        if len(mdata) == len(data):
            tempname = f'temp_{orb}_{str(random.randint(0, 1000000))}'
            field = 'orbit'
            quant = orb
        else:
            tempname = f'temp_query_{str(random.randint(0, 1000000))}'
            field = 'query'
            quant = data.query
    else:
        tempname = f'temp_query_{str(random.randint(0, 1000000))}'
        field = 'query'
        quant = data.query
        
    return tempname, field, quant


class LOSResult(ModelResult):
    '''Class to contain the LOS result from multiple outputfiles.'''
    def __init__(self, start_from, mdata, quantity, dphi=3*u.deg,
                 filenames=None, overwrite=False, fit_to_data=False, **kwargs):
        """Determine column or emission along lines of sight.
        This assumes the model has already been run.
        
        Parameters
        ==========
        start_from
            Either an Input or Output object
        
        data
            A Pandas DataFrame object with information on the lines of sight.
            
        quantity
            Quantity to calculate: 'column', 'radiance', 'density'
            
        dphi
            Angular size of the view cone. Default = 3 deg.
            
        filenames
            A filename or list of filenames to use. Default = None is to
            find all files created for the inputs. Not used when start_from
            is an Output object.
            
        overwrite
            If True, deletes any images that have already been computed. Not
            used when start_from is an Output object.
            Default = False
        """
        format_ = {'quantity':quantity}
        if isinstance(start_from, Input):
            # Given inputs: Will need to search for and/or compute outputs
            inputs = start_from
            output = None
            super().__init__(inputs, format_, filenames=filenames)
        elif isinstance(start_from, Output):
            # Starting from an precomputed output
            output = start_from
            inputs = output.inputs
            super().__init__(inputs, format_, output=output)
        else:
            raise TypeError
        
        # Make sure data is in the right reference frame
        mdata.set_frame('Model')
        data = mdata.data

        tstart = datetime.now()
        self.type = 'LineOfSight'
        self.species = inputs.options.species
        self.origin = inputs.geometry.planet
        self.unit = u.def_unit('R_' + self.origin.object,
                               self.origin.radius)
        self.dphi = dphi.to(u.rad).value
        # Just compute lines of sight or also try to fit
        self.fitted = fit_to_data

        nspec = len(data)
        self.radiance = np.zeros(nspec)
        self.packets = pd.DataFrame()
        self.ninview = np.zeros(nspec, dtype=int)
        saved_files = {}
        for j,outfile in enumerate(self.filenames):
            if outfile is None:
                # Output was provided - outfile not known
                if self.fitted:
                    # Fit the outputs to the data
                    masking = kwargs['masking'] if 'masking' in kwargs else None
                    radiance_, npackets_, weighting, packets = self.fit_model_to_data(
                            data, output, masking=masking)
                else:
                    # Just compute lines of sight
                    radiance_, npackets_ = self.create_model(data, output)
                    weighting, packets  = None, None
                print(f'Completed model {j+1} of {len(self.filenames)}')
                del output
            else:
                # Search to see if it is already done
                restored = self.restore(data, outfile)
                radiance_, npackets_, idnum, weighting, packets, saved_file = restored
                if (radiance_ is None) or overwrite:
                    if (radiance_ is not None) and overwrite:
                        self.delete_model(idnum, data)
                    else:
                        pass

                    output = Output.restore(outfile)
                    if self.fitted:
                        masking = kwargs['masking'] if 'masking' in kwargs else None
                        radiance_, npackets_, weighting_, packets = self.fit_model_to_data(
                                data, output, masking=masking)
                        saved_file = self.save(data, outfile, radiance_,
                                               npackets_, weighting_, packets)
                    else:
                        radiance_, npackets_ = self.create_model(data, output)
                        saved_file = self.save(data, outfile, radiance_,
                                               npackets_)
                    saved_files[outfile] = saved_file
                    print(f'Completed model {j+1} of {len(self.filenames)}')
                    del output
                else:
                    saved_files[outfile] = saved_file
                    print(f'Model {j+1} of {len(self.filenames)} '
                          'previously completed.')

            self.radiance += radiance_.values
            self.ninview += npackets_.astype(int).values
            if weighting is not None:
                self.weighting = weighting
                self.packets = packets
            else:
                pass

        self.savefiles = saved_files
        self.radiance *= u.R
        tend = datetime.now()
        print(f'Total time = {tend-tstart}')

    def delete_model(self, data, idnum):
        _, field, _ = determine_model_type(data)
    
        with database_connect() as con:
            cur = con.cursor()
            cur.execute(f'''SELECT idnum, filename FROM uvvsmodels_{field}
                           WHERE out_idnum = %s''', (int(idnum), ))
            assert cur.rowcount in (0, 1)
            for mid, mfile in cur.fetchall():
                cur.execute(f'''DELETE from uvvsmodels_{field}
                               WHERE idnum = %s''', (mid, ))
                if os.path.exists(mfile):
                    os.remove(mfile)

    def save(self, data, fname, radiance, npackets,
             weighting=None, packets=None):
        with database_connect() as con:
            cur = con.cursor()

            # Determine the id of the outputfile
            idnum_ = pd.read_sql(f'''SELECT idnum
                                    FROM outputfile
                                    WHERE filename='{fname}' ''', con)
            idnum = int(idnum_.idnum[0])

            # Insert the model into the database
            if self.quantity == 'radiance':
                mech = ', '.join(sorted([m for m in self.mechanism]))
                wave_ = sorted([w.value for w in self.wavelength])
                wave = ', '.join([str(w) for w in wave_])
            else:
                mech = None
                wave = None

            # Determine if model represents a single complete orbit
            tempname, field, quant = determine_model_type(data)

            cur.execute(f'''INSERT into uvvsmodels_{field} (out_idnum, quantity,
                            {field}, dphi, mechanism, wavelength,
                            fitted, filename)
                            values (%s, %s, %s, %s, %s, %s, %s, %s)''',
                        (idnum, self.quantity, quant, self.dphi,
                         mech, wave, self.fitted, tempname))

            # Determine the savefile name
            idnum_ = pd.read_sql(f'''SELECT idnum
                                     FROM uvvsmodels_{field}
                                     WHERE filename='{tempname}';''', con)
            assert len(idnum_) == 1
            idnum = int(idnum_.idnum[0])

            savefile = os.path.join(os.path.dirname(fname),
                                    f'model.{idnum}.pkl')
            if self.fitted:
                with open(savefile, 'wb') as f:
                    pickle.dump((radiance, npackets, weighting, packets), f)
            else:
                with open(savefile, 'wb') as f:
                    pickle.dump((radiance, npackets), f)
            cur.execute(f'''UPDATE uvvsmodels_{field}
                            SET filename=%s
                            WHERE idnum=%s''', (savefile, idnum))
    
        return savefile

    def restore(self, data, fname):
        tempname, field, quant = determine_model_type(data)
        
        with database_connect() as con:
            # Determine the id of the outputfile
            idnum_ = pd.read_sql(f'''SELECT idnum
                                    FROM outputfile
                                    WHERE filename='{fname}' ''', con)
            oid = idnum_.idnum[0]

            if self.quantity == 'radiance':
                mech = ("mechanism = '" +
                        ", ".join(sorted([m for m in self.mechanism])) +
                        "'")
                wave_ = sorted([w.value for w in self.wavelength])
                wave = ("wavelength = '" +
                        ", ".join([str(w) for w in wave_]) +
                        "'")
            else:
                mech = 'mechanism is NULL'
                wave = 'wavelength is NULL'

            if field == 'query':
                quant = f"'{quant}'"
            else:
                pass
            result = pd.read_sql(
                f'''SELECT idnum, filename FROM uvvsmodels_{field}
                    WHERE out_idnum={oid} and
                          quantity = '{self.quantity}' and
                          {field} = {quant} and
                          dphi = {self.dphi} and
                          {mech} and
                          {wave} and
                          fitted = {self.fitted}''', con)

            assert len(result) <= 1
            if len(result) == 1:
                savefile = result.filename[0]

                if self.fitted:
                    with open(savefile, 'rb') as f:
                        radiance, npackets, weighting, packets = pickle.load(f)
                else:
                    with open(savefile, 'rb') as f:
                        radiance, npackets = pickle.load(f)
                    weighting, packets = None, None
                idnum = result.idnum[0]
                
                if len(radiance) != len(data):
                    radiance, npackets, idnum, weighting = None, None, None, None
                else:
                    pass
            else:
                savefile = None
                radiance, npackets, idnum = None, None, None
                weighting, packets = None, None

        return radiance, npackets, idnum, weighting, packets, savefile
    
    def create_model(self, data, output):
        # distance of s/c from planet
        dist_from_plan = np.sqrt(data.x**2 + data.y**2 + data.z**2)

        # Angle between look direction and planet.
        ang = np.arccos((-data.x*data.xbore - data.y*data.ybore -
                         data.z*data.zbore)/dist_from_plan)

        # Check to see if look direction intersects the planet anywhere
        asize_plan = np.arcsin(1./dist_from_plan)

        # Don't worry about lines of sight that don't hit the planet
        dist_from_plan.loc[ang > asize_plan] = 1e30

        # Load the outputfile
        packets = output.X
        packets['radvel_sun'] = (packets['vy'] +
                                 output.vrplanet.to(self.unit/u.s).value)

        # Will base shadow on line of sight, not the packets
        out_of_shadow = np.ones(len(packets))
        self.packet_weighting(packets, out_of_shadow, output.aplanet)

        xcols = ['x', 'y', 'z']
        borecols = ['xbore', 'ybore', 'zbore']

        # This sets limits on regions where packets might be
        tree = KDTree(packets[xcols].values)
        # tree = BallTree(packets[xcols].values)

        # This removes the packets that aren't close to the los
        oedge = output.inputs.options.outeredge*2
        
        # This sets limits on regions where packets might be
        rad = pd.Series(np.zeros(data.shape[0]), index=data.index)
        npack = pd.Series(np.zeros(data.shape[0]), index=data.index)

        print(f'{data.shape[0]} spectra taken.')
        for i, spectrum in data.iterrows():
            x_sc = spectrum[xcols].values.astype(float)
            bore = spectrum[borecols].values.astype(float)

            dd = 30  # Furthest distance we need to look
            x_far = x_sc+bore*dd
            while np.linalg.norm(x_far) > oedge:
                dd -= 0.1
                x_far = x_sc+bore*dd
                
            t = [0.05]
            while t[-1] < dd:
                t.append(t[-1] + t[-1]*np.sin(self.dphi))
            t = np.array(t)
            Xbore = x_sc[np.newaxis, :]+bore[np.newaxis, :]*t[:, np.newaxis]

            wid = t*np.sin(self.dphi)
            ind = np.concatenate(tree.query_radius(Xbore, wid))
            ilocs = np.unique(ind).astype(int)
            indicies = packets.iloc[ilocs].index
            subset = packets.loc[indicies]

            xpr = subset[xcols]-x_sc[np.newaxis, :]
            rpr = np.sqrt(xpr['x']*xpr['x']+
                          xpr['y']*xpr['y']+
                          xpr['z']*xpr['z'])

            losrad = np.sum(xpr*bore[np.newaxis, :], axis=1)
            inview = rpr < dist_from_plan[i]
            
            if np.any(inview):
                Apix = np.pi*(rpr[inview]*np.sin(self.dphi))**2*(
                    self.unit.to(u.cm))**2
                wtemp = subset.loc[inview, 'weight']/Apix*self.atoms_per_packet
                if self.quantity == 'radiance':
                    # Determine if any packets are in shadow
                    # Projection of packet onto LOS
                    # Point along LOS the packet represents
                    losrad_ = losrad[inview].values
                    hit = (x_sc[np.newaxis,:] +
                           bore[np.newaxis,:] * losrad_[:,np.newaxis])
                    rhohit = np.linalg.norm(hit[:,[0,2]], axis=1)
                    out_of_shadow = (rhohit > 1) | (hit[:,1] < 0)
                    wtemp *= out_of_shadow

                    rad.loc[i] = wtemp.sum()
                    npack.loc[i] = sum(inview)
                else:
                    assert False, 'Other quantities not set up.'
            else:
                pass

            if len(data) > 10:
                ind = data.index.get_loc(i)
                if (ind%(len(data)//10)) == 0:
                    print(f'Completed {ind+1} spectra')

        return rad, npack
    
    def fit_model_to_data(self, data, output, masking=None, makeplots=False):

        def should_add_weight(index, saved):
            return index in saved

        def add_weight(x, ratio):
            return np.append(x, ratio)
        
        def add_index(x, i):
            return np.append(x, i)

        # distance of s/c from planet
        dist_from_plan = np.sqrt(data.x**2+data.y**2+data.z**2)

        # Angle between look direction and planet.
        ang = np.arccos((-data.x*data.xbore-data.y*data.ybore-
                         data.z*data.zbore)/dist_from_plan)

        # Check to see if look direction intersects the planet anywhere
        asize_plan = np.arcsin(1./dist_from_plan)

        # Don't worry about lines of sight that don't hit the planet
        dist_from_plan.loc[ang > asize_plan] = 1e30

        # Load the outputfile
        packets = output.X
        packets['radvel_sun'] = (packets['vy']+
                                 output.vrplanet.to(self.unit/u.s).value)
    
        # Will base shadow on line of sight, not the packets
        out_of_shadow = np.ones(packets.shape[0])
        self.packet_weighting(packets, out_of_shadow, output.aplanet)
    
        xcols = ['x', 'y', 'z']
        borecols = ['xbore', 'ybore', 'zbore']
    
        # This sets limits on regions where packets might be
        tree = KDTree(packets[xcols].values)
        # tree = BallTree(packets[xcols].values)
    
        # Maxiumum line of sight length
        oedge = output.inputs.options.outeredge*2
    
        # rad = modeled radiance
        # npack = number of packets used to compute radiance
        # saved_packets = list of indicies of the packets used for each spectrum
        # weighting = list of the weights that should be applied
        #   - Final weighting for each packet is mean of weights
        rad = pd.Series(np.zeros(data.shape[0]), index=data.index)
        npack = pd.Series(np.zeros(data.shape[0]), index=data.index)
        saved_packets = pd.Series((np.ndarray((0,), dtype=int)
                                   for _ in range(data.shape[0])),
                                  index=data.index)
        ind0 = packets.Index.unique()
        weighting = pd.Series((np.ndarray((0,)) for _ in range(ind0.shape[0])),
                              index=ind0)
        included = pd.Series((np.ndarray((0,), dtype=np.int)
                              for _ in range(ind0.shape[0])), index=ind0)

        # Determine which points should be used for the fit
        _, _, mask = fit_model(data.radiance, None, data.sigma,
                               masking=masking, mask_only=True,
                               altitude=data.alttan)
    
        print(f'{data.shape[0]} spectra taken.')
        for i, spectrum in data.iterrows():
            x_sc = spectrum[xcols].values.astype(float)
            bore = spectrum[borecols].values.astype(float)
        
            dd = 30  # Furthest distance we need to look
            x_far = x_sc+bore*dd
            while np.linalg.norm(x_far) > oedge:
                dd -= 0.1
                x_far = x_sc+bore*dd
        
            t = [0.05]
            while t[-1] < dd:
                t.append(t[-1]+t[-1]*np.sin(self.dphi))
            t = np.array(t)
            Xbore = x_sc[np.newaxis, :]+bore[np.newaxis, :]*t[:, np.newaxis]
        
            wid = t*np.sin(self.dphi)
            ind = np.concatenate(tree.query_radius(Xbore, wid))
            ilocs = np.unique(ind).astype(int)
            indicies = packets.iloc[ilocs].index
            subset = packets.loc[indicies]
        
            xpr = subset[xcols]-x_sc[np.newaxis, :]
            rpr = np.sqrt(xpr['x']*xpr['x']+
                          xpr['y']*xpr['y']+
                          xpr['z']*xpr['z'])
    
            losrad = np.sum(xpr*bore[np.newaxis, :], axis=1)
            inview = rpr < dist_from_plan[i]
        
            if np.any(inview):
                Apix = np.pi*(rpr[inview]*np.sin(self.dphi))**2*(
                    self.unit.to(u.cm))**2
                wtemp = subset.loc[inview, 'weight']/Apix*self.atoms_per_packet
                if self.quantity == 'radiance':
                    # Determine if any packets are in shadow
                    # Projection of packet onto LOS
                    # Point along LOS the packet represents
                    losrad_ = losrad[inview].values
                    hit = (x_sc[np.newaxis, :]+
                           bore[np.newaxis, :]*losrad_[:, np.newaxis])
                    rhohit = np.linalg.norm(hit[:, [0, 2]], axis=1)
                    out_of_shadow = (rhohit > 1) | (hit[:, 1] < 0)
                    wtemp *= out_of_shadow
                
                    rad.loc[i] = wtemp.sum()
                    npack.loc[i] = sum(inview)
                    if (wtemp.sum() > 0) and mask[i]:
                        ratio = spectrum.radiance/wtemp.sum()
                
                        # Save which packets are used for each spectrum
                        saved_packets[i] = subset.loc[inview, 'Index'].unique()
                        should = weighting.index.to_series().apply(
                                should_add_weight, args=(saved_packets[i],))
                        weighting.loc[should] = weighting.loc[should].apply(
                                add_weight, args=(ratio,))
                        included.loc[should] = included.loc[should].apply(
                                add_index, args=(i,))
                    else:
                        pass
                else:
                    assert False, 'Other quantities not set up.'
            else:
                pass
        
            if len(data) > 10:
                ind = data.index.get_loc(i)
                if (ind%(len(data)//10)) == 0:
                    print(f'Completed {ind+1} spectra')

        # Determine the proper weightings
        new_weight = weighting.apply(lambda x: x.mean() if x.shape[0] > 0 else 0.)
        assert np.all(np.isfinite(new_weight))
        if np.any(new_weight > 0):
            multiplier = new_weight.loc[packets['Index']].values
            output.X.loc[:, 'frac'] = packets.loc[:, 'frac']*multiplier
            output.X0.loc[:, 'frac'] = output.X0.loc[:, 'frac']*new_weight
            output.totalsource = np.sum(output.X0.frac)
            
            rad, npack = self.create_model(data, output)
        else:
            rad = pd.Series(np.zeros(data.shape[0]), index=data.index)
            npack = pd.Series(np.zeros(data.shape[0]), index=data.index)

        assert np.all(weighting.apply(len) == included.apply(len))
        weighting = pd.DataFrame({'weight': weighting.values,
                                  'included': included.values})

        return rad, npack, weighting, saved_packets



import os.path
import numpy as np
import pandas as pd
import pickle
import random
import astropy.units as u
from datetime import datetime
from sklearn.neighbors import KDTree
#from scipy.spatial import KDTree, cKDTree
from .ModelResults import ModelResult
from .database_connect import database_connect
from .Input import Input
from .Output import Output


class LOSResult(ModelResult):
    def __init__(self, start_from, data, quantity, dphi=3*u.deg,
                 filenames=None, overwrite=False, savepackets=False,
                 **kwargs):
        """Determine column or emission along lines of sight.
        This assumes the model has already been run.
        
        Parameters
        ==========
        start_from
            Either an Input or Output object
        
        data
            A Pandas DataFrame object with information on the lines of sight.
            
        quantity
            Quantity to calculate: 'column', 'radiance', 'density'
            
        dphi
            Angular size of the view cone. Default = 3 deg.
            
        filenames
            A filename or list of filenames to use. Default = None is to
            find all files created for the inputs. Not used when start_from
            is an Output object.
            
        overwrite
            If True, deletes any images that have already been computed. Not
            used when start_from is an Output object.
            Default = False
        """
        format_ = {'quantity':quantity}
        if isinstance(start_from, Input):
            inputs = start_from
            super().__init__(inputs, format_, filenames=filenames)
        elif isinstance(start_from, Output):
            output = start_from
            inputs = output.inputs
            super().__init__(inputs, format_, output=output)
        else:
            raise TypeError

        tstart = datetime.now()
        self.type = 'LineOfSight'
        self.species = inputs.options.species
        self.origin = inputs.geometry.planet
        self.unit = u.def_unit('R_' + self.origin.object,
                               self.origin.radius)
        self.dphi = dphi.to(u.rad).value

        nspec = len(data)
        self.radiance = np.zeros(nspec)
        self.packets = pd.DataFrame()
        self.ninview = np.zeros(nspec, dtype=int)
        for j,outfile in enumerate(self.filenames):
            if outfile is None:
                radiance_, npackets_, saved_ = self.create_model(data, output,
                                                    savepackets=savepackets)
                print(f'Completed model {j+1} of {len(self.filenames)}')
            else:
                # Search to see if it is already done
                radiance_, npackets_, saved_, idnum = self.restore(data, outfile)

                if (radiance_ is None) or overwrite:
                    if (radiance_ is not None) and overwrite:
                        self.delete_model(idnum)
                    else:
                        pass

                    output = Output.restore(outfile)
                    radiance_, npackets_, saved_ = self.create_model(data,
                                            output, savepackets=savepackets)
                    self.save(data, outfile, radiance_, npackets_, saved_)
                    print(f'Completed model {j+1} of {len(self.filenames)}')
                else:
                    print(f'Model {j+1} of {len(self.filenames)} '
                          'previously completed.')

            self.radiance += radiance_
            self.ninview += npackets_.astype(int)
            self.packets[j] = saved_

        self.radiance *= self.atoms_per_packet
        self.radiance *= u.R
        tend = datetime.now()
        print(f'Total time = {tend-tstart}')

    def delete_model(self, idnum):
        with database_connect() as con:
            cur = con.cursor()
            cur.execute('''SELECT idnum, filename FROM uvvsmodels
                           WHERE out_idnum = %s''', (int(idnum), ))
            assert cur.rowcount in (0, 1)
            for mid, mfile in cur.fetchall():
                cur.execute('''DELETE from uvvsmodels
                               WHERE idnum = %s''', (mid, ))
                if os.path.exists(mfile):
                    os.remove(mfile)

    def save(self, data, fname, radiance, packets, saved_packets):
        # Determine if the model can be saved.
        # Criteria: 1 complete orbit, nothing more.
        orbits = set(data.orbit)
        orb = orbits.pop()

        if len(orbits) != 0:
            print('Model spans more than one orbit. Cannot be saved.')
        else:
            from MESSENGERuvvs import MESSENGERdata
            mdata = MESSENGERdata(self.species, f'orbit = {orb}')
            if len(mdata) != len(data):
                print('Model does not contain the complete orbit. '
                      'Cannot be saved.')
            else:
                with database_connect() as con:
                    cur = con.cursor()

                    # Determine the id of the outputfile
                    idnum_ = pd.read_sql(f'''SELECT idnum
                                            FROM outputfile
                                            WHERE filename='{fname}' ''', con)
                    idnum = int(idnum_.idnum[0])

                    # Insert the model into the database
                    if self.quantity == 'radiance':
                        mech = ', '.join(sorted([m for m in self.mechanism]))
                        wave_ = sorted([w.value for w in self.wavelength])
                        wave = ', '.join([str(w) for w in wave_])
                    else:
                        mech = None
                        wave = None

                    tempname = f'temp_{orb}_{str(random.randint(0, 1000000))}'
                    cur.execute(f'''INSERT into uvvsmodels (out_idnum, quantity,
                                    orbit, dphi, mechanism, wavelength, filename)
                                    values (%s, %s, %s, %s, %s, %s, %s)''',
                                (idnum, self.quantity, orb, self.dphi,
                                 mech, wave, tempname))

                    # Determine the savefile name
                    idnum_ = pd.read_sql(f'''SELECT idnum
                                             FROM uvvsmodels
                                             WHERE filename='{tempname}';''', con)
                    assert len(idnum_) == 1
                    idnum = int(idnum_.idnum[0])

                    savefile = os.path.join(os.path.dirname(fname),
                                        f'model.orbit{orb:04}.{idnum}.pkl')
                    with open(savefile, 'wb') as f:
                        pickle.dump((radiance, packets, saved_packets), f)
                    cur.execute(f'''UPDATE uvvsmodels
                                    SET filename=%s
                                    WHERE idnum=%s''', (savefile, idnum))

    def restore(self, data, fname):
        # Determine if the model can be restored.
        # Criteria: 1 complete orbit, nothing more.
        orbits = set(data.orbit)
        orb = orbits.pop()

        if len(orbits) != 0:
            print('Model spans more than one orbit. Cannot be saved.')
            radiance, packets, idnum = None, None, None
        else:
            with database_connect() as con:
                # Determine the id of the outputfile
                idnum_ = pd.read_sql(f'''SELECT idnum
                                        FROM outputfile
                                        WHERE filename='{fname}' ''', con)
                oid = idnum_.idnum[0]

                if self.quantity == 'radiance':
                    mech = ("mechanism = '" +
                            ", ".join(sorted([m for m in self.mechanism])) +
                            "'")
                    wave_ = sorted([w.value for w in self.wavelength])
                    wave = ("wavelength = '" +
                            ", ".join([str(w) for w in wave_]) +
                            "'")
                else:
                    mech = 'mechanism is NULL'
                    wave = 'wavelength is NULL'

                result = pd.read_sql(
                    f'''SELECT idnum, filename FROM uvvsmodels
                        WHERE out_idnum={oid} and
                              quantity = '{self.quantity}' and
                              orbit = {orb} and
                              dphi = {self.dphi} and
                              {mech} and
                              {wave}''', con)

            assert len(result) <= 1
            if len(result) == 1:
                savefile = result.filename[0]

                with open(savefile, 'rb') as f:
                    radiance, packets, saved_ = pickle.load(f)
                idnum = result.idnum[0]
                if len(radiance) != len(data):
                    radiance, packets, saved_, idnum = None, None, None, None
                else:
                    pass
            else:
                radiance, packets, saved_, idnum = None, None, None, None

        return radiance, packets, saved_, idnum

    def create_model(self, data, output, savepackets, **kwargs):
        # distance of s/c from planet
        dist_from_plan = (np.sqrt(data.x**2 + data.y**2 + data.z**2)).values

        # Angle between look direction and planet.
        ang = np.arccos((-data.x*data.xbore - data.y*data.ybore -
                         data.z*data.zbore)/dist_from_plan)

        # Check to see if look direction intersects the planet anywhere
        asize_plan = np.arcsin(1./dist_from_plan)

        # Don't worry about lines of sight that don't hit the planet
        dist_from_plan[ang > asize_plan] = 1e30

        # Load the outputfile
        packets = output.X
        packets['radvel_sun'] = (packets['vy'] +
                                 output.vrplanet.to(self.unit/u.s).value)

        # Will base shadow on line of sight, not the packets
        out_of_shadow = np.ones(len(packets))
        self.packet_weighting(packets, out_of_shadow, output.aplanet)
        
        xpack = packets[['x', 'y', 'z']].values
        weight = packets['weight'].values
        index = packets['index'].values
        tree = KDTree(xpack)
        # tree = BallTree(xpack)

        # This sets limits on regions where packets might be
        rad, pack = np.zeros(data.shape[0]), np.zeros(data.shape[0])
        if savepackets:
            saved_packets = np.ndarray(data.shape[0], dtype='O')
        else:
            saved_packets = None

        xdata = data[['x', 'y', 'z']].values.astype(float)
        boresight = data[['xbore', 'ybore', 'zbore']].values.astype(float)
        
        # This removes the packets that aren't close to the los
        if 'outeredge' in kwargs:
            oedge = kwargs['outeredge']
        else:
            oedge = output.inputs.options.outeredge*2

        print(f'{data.shape[0]} spectra taken.')
        for i in range(data.shape[0]):
            x_sc = xdata[i,:]
            bore = boresight[i,:]

            dd = 30  # Furthest distance we need to look
            x_far = x_sc+bore*dd
            while np.linalg.norm(x_far) > oedge:
                dd -= 0.1
                x_far = x_sc+bore*dd
                
            t = [0.05]
            while t[-1] < dd:
                t.append(t[-1] + t[-1]*np.sin(self.dphi))
            t = np.array(t)
            Xbore = x_sc[np.newaxis, :]+bore[np.newaxis, :]*t[:, np.newaxis]

            wid = t*np.sin(self.dphi)
            ind = np.concatenate(tree.query_radius(Xbore, wid))
            indicies = np.unique(ind).astype(int)
            subset = xpack[indicies, :]
            wt = weight[indicies]
            subindex = index[indicies]

            xpr = subset-x_sc[np.newaxis, :]
            rpr = np.sqrt(xpr[:, 0]*xpr[:, 0]+xpr[:, 1]*xpr[:, 1]+
                          xpr[:, 2]*xpr[:, 2])
            losrad = np.sum(xpr*bore[np.newaxis, :], axis=1)
            inview = rpr < dist_from_plan[i]
            
            if np.any(inview):
                Apix = np.pi*(rpr[inview]*np.sin(self.dphi))**2 * self.unit**2
                wtemp = wt[inview]/Apix.to(u.cm**2).value
                if self.quantity == 'radiance':
                    losrad_ = losrad[inview]
                    # Determine if any packets are in shadow
                    # Projection of packet onto LOS
                    # Point along LOS the packet represents
                    hit = (x_sc[np.newaxis,:] +
                           bore[np.newaxis,:] * losrad_[:,np.newaxis])
                    rhohit = np.linalg.norm(hit[:,[0,2]], axis=1)
                    out_of_shadow = (rhohit > 1) | (hit[:,1] < 0)
                    wtemp *= out_of_shadow

                    rad[i] = np.sum(wtemp)
                    pack[i] = len(losrad_)
                    if savepackets:
                        saved_packets[i] = tuple(set(subindex[inview]))
                    else:
                        pass
            else:
                pass

            if len(data) > 10:
                if (i%(len(data)//10)) == 0:
                    print(f'Completed {i+1} spectra')
                    
        del output
        return rad, pack, saved_packets



"""Create and read configuration file, create necessary database tables."""
import os
import os.path
from .database_connect import database_connect
import sys
import types


def configfile():
    """Configure external resources used in the model.

    The following parameters can be saved in the file `$HOME/.nexoclom`.
    
    * savepath = <path where output files are saved>
    
    * datapath = <path where MESSENGER data is kept>
    
    * database = <name of the postgresql database to use> (*optional*)
    
    * port = <port for postgreSQL server to use> (*optional*)
    
    If savepath and datapath are not present, user is prompted to enter them.
    """
    # Determne the savepath
    cfile = os.path.join(os.environ['HOME'], '.nexoclom')
    savepath = None
    if os.path.isfile(cfile):
        for line in open(cfile, 'r').readlines():
            key, value = line.split('=')
            if key.strip() == 'savepath':
                savepath = value.strip()
            else:
                pass
    else:
        pass

    if savepath is None:
        savepath = input('Where should outputs be saved: ')
        with open(cfile, 'a') as f:
            f.write(f'savepath = {savepath}\n')

        # Create save directory if necessary
        if not os.path.isdir(savepath):
            try:
                os.makedirs(savepath)
            except:
                assert 0, f'Could not create directory {savepath}'

    # Create the database if necessary
    database, port = database_connect(return_con=False)
    with database_connect(database='postgres') as con:
        cur = con.cursor()
        cur.execute('select datname from pg_database')
        dbs = [r[0] for r in cur.fetchall()]

        if database not in dbs:
            print(f'Creating database {database}')
            cur.execute(f'create database {database}')
        else:
            pass

    return savepath


def verify_output_tables():
    """Create the database tables used by nexoclom to save output."""
    with database_connect() as con:
        cur = con.cursor()
        cur.execute('select table_name from information_schema.tables')
    tables = [r[0] for r in cur.fetchall()]
    
    with open(os.path.join(os.path.dirname(__file__),
                           'data',
                           'schema.sql'), 'r') as sqlfile:
        done = False
        while not done:
            line = sqlfile.readline()
            nextline = ''
            if 'TABLE' in line:
                table_to_test = line[len('CREATE TABLE '):-3]
                if table_to_test in tables:
                    # Need to verify schema
                    pass
                else:
                    query = line
                    nextline = sqlfile.readline()
                    while (nextline.strip()) and ('DONE' not in nextline):
                        query += nextline
                        nextline = sqlfile.readline()
                    print(query)
                    cur.execute(query)
            done = ('DONE' in nextline) or ('DONE' in line)


def configure_model():
    """Ensure the database and configuration file are set up for nexoclom.
    
    **Parameters**
    
    No parameters.
    
    **Returns**
    
    No output.
    """
    if isinstance(sys.modules['psycopg2'], types.ModuleType):
        # Get database name and port
        database, port = database_connect(return_con=False)

        # Verify database is running
        status = os.popen('pg_ctl status').read()
        if 'no server running' in status:
            os.system(f'pg_ctl start -D $HOME/.postgres/main '
                      f'-l $HOME/.postgres/logfile -o "-p {port}"')
        else:
            pass
        
        with database_connect() as con:
            # Check whether SSObject has been created
            cur = con.cursor()
            query = """
                    SELECT exists
                    (select 1 from pg_type where typname = 'ssobject');
                    """
            cur.execute(query)
            result = cur.fetchall()

        if not result[0][0]:
            from solarsystemMB import create_SSObject
            create_SSObject()

        else:
            pass
        
        verify_output_tables()



import os.path
import numpy as np
import pandas as pd
import astropy.units as u
import mathMB
from atomicdataMB import gValue
from .database_connect import database_connect
from .input_classes import InputError
from .Output import Output


class ModelResult:
    def __init__(self, inputs, format, filenames=None, output=None):
        self.inputs = inputs
        if isinstance(output, Output):
            # Output is given
            self.npackets = output.npackets
            self.totalsource = output.totalsource
            self.filenames = [None]
        elif isinstance(filenames, str):
            self.filenames = [filenames]
            with database_connect() as con:
                packs = pd.read_sql(f'''SELECT npackets, totalsource
                                        FROM outputfile
                                        WHERE filename={filenames}''', con)
            self.npackets = packs.npackets[0]
            self.totalsource = packs.totalsource[0]
        elif filenames is None:
            self.filenames, self.npackets, self.totalsource = inputs.search()
        else:
            raise Exception
            
        if self.npackets == 0:
            pass
        else:
            self.mod_rate = self.totalsource/inputs.options.endtime.value
            self.atoms_per_packet = 1e23/self.mod_rate
            print(f'Total number of packets run = {self.npackets}')
            print(f'Total source = {self.totalsource} packets')
            print(f'1 packet represents {self.atoms_per_packet} atoms')
            print(f'Model rate = {self.mod_rate} packets/sec')

            if isinstance(format, str):
                if os.path.exists(format):
                    self.format = {}
                    with open(format, 'r') as f:
                        for line in f:
                            if ';' in line:
                                line = line[:line.find(';')]
                            elif '#' in line:
                                line = line[:line.find('#')]
                            else:
                                pass
                            
                            if '=' in line:
                                p, v = line.split('=')
                                self.format[p.strip().lower()] = v.strip()
                            else:
                                pass
                else:
                    raise FileNotFoundError('ModelResult.__init__',
                                            'Format file not found.')
            elif isinstance(format, dict):
                self.format = format
            else:
                raise TypeError('ModelResult.__init__',
                                'format must be a dict or filename.')
            
            # Do some validation
            quantities = ['column', 'radiance', 'density']

            if 'quantity' in self.format:
                if self.format['quantity'] in quantities:
                    self.quantity = self.format['quantity']
                else:
                    raise InputError('ModelImage.__init__',
                                     "quantity must be 'column' or 'radiance'")
            else:
                raise InputError('ModelImage.__init__',
                                 'quantity must be specified.')

            if self.quantity == 'radiance':
                # Note - only resonant scattering currently possible
                self.mechanism = ['resonant scattering']
        
                if 'wavelength' in self.format:
                    self.wavelength = tuple(
                        int(m.strip())*u.AA
                        for m
                        in self.format['wavelength'].split(','))
                elif inputs.options.species == 'Na':
                    self.wavelength = (5891*u.AA, 5897*u.AA)
                elif inputs.options.species == 'Ca':
                    self.wavelength = (4227*u.AA,)
                elif inputs.options.species == 'Mg':
                    self.wavelength = (2852*u.AA,)
                else:
                    raise InputError('ModelResult.__init__', ('Default wavelengths '
                                  f'not available for {inputs.options.species}'))
            else:
                pass

    def transform_reference_frame(self, output):
        """If the image center is not the planet, transform to a
           moon-centric reference frame."""
        assert 0, 'Not ready yet.'

        # Load output

        # # Transform to moon-centric frame if necessary
        # if result.origin != result.inputs.geometry.planet:
        #     assert 0, 'Need to do transformation for a moon.'
        # else:
        #     origin = np.array([0., 0., 0.])*output.x.unit
        #     sc = 1.

        # Choose which packets to use
        # touse = output.frac >= 0 if keepall else output.frac > 0

        # packet positions relative to origin -- not rotated
        # pts_sun = np.array((output.x[touse]-origin[0],
        #                     output.y[touse]-origin[1],
        #                     output.z[touse]-origin[2]))*output.x.unit
        #
        # # Velocities relative to sun
        # vels_sun = np.array((output.vx[touse],
        #                      output.vy[touse],
        #                      output.vz[touse]))*output.vx.unit

        # Fractional content
        # frac = output.frac[touse]

        return output #, pts_sun, vels_sun, frac

    def packet_weighting(self, packets, out_of_shadow, aplanet):
        if self.quantity == 'column':
            packets['weight'] = packets['frac']
        elif self.quantity == 'density':
            packets['weight'] = packets['frac']
        elif self.quantity == 'radiance':
            if 'resonant scattering' in self.mechanism:
                gg = np.zeros(len(packets))/u.s
                for w in self.wavelength:
                    gval = gValue(self.inputs.options.species, w, aplanet)
                    gg += mathMB.interpu(packets['radvel_sun'].values *
                                         self.unit/u.s, gval.velocity, gval.g)

                weight_resscat = packets['frac']*out_of_shadow*gg.value/1e6
            else:
                weight_resscat = np.zeros(len(packets))
                
            packets['weight'] = weight_resscat # + other stuff

        assert np.all(np.isfinite(packets['weight'])), 'Non-finite weights'



import os
import psycopg2


def database_connect(database=None, port=None, return_con=True):
    """Wrapper for psycopg2.connect() that determines which database and port to use.

    :return:
    :param database: Default = None to use value from config file
    :param port: Default = None to use value from config file
    :param return_con: False to return database name and port instead of connection
    :return: Database connection with autocommit = True unless return_con = False
    """
    configfile = os.path.join(os.environ['HOME'], '.nexoclom')
    config = {}
    if os.path.isfile(configfile):
        for line in open(configfile, 'r').readlines():
            key, value = line.split('=')
            config[key.strip()] = value.strip()

        if (database is None) and ('database' in config):
            database = config['database']
        else:
            pass

        if (port is None) and ('port' in config):
            port = int(config['port'])
        else:
            pass
    else:
        pass

    if database is None:
        database = 'thesolarsystemmb'
    else:
        pass

    if port is None:
        port = 5432
    else:
        pass

    if return_con:
        con = psycopg2.connect(database=database, port=port)
        con.autocommit = True

        return con
    else:
        return database, port



import numpy as np
import os.path
from scipy.io import readsav
try:
    from .SurfaceInteraction import SurfaceInteraction
    from .Input import Input
except:
    from nexoclom.SurfaceInteraction import SurfaceInteraction
    from nexoclom import Input

def test_SurfaceInteraction():
    idlout = readsav(os.path.join(os.path.dirname(__file__),
                                  'data', 'emit.sav'))
    emit_idl = idlout['emit']
    vgrid_idl = emit_idl['vgrid'][0].transpose()
    prob_idl = emit_idl['probability'][0]
    temp_idl = emit_idl['temperature'][0]


    inputs = Input('inputfiles/Na.maxwellian.1200.accom.input')
    emit_py = SurfaceInteraction(inputs, nt=21, nv=101, nprob=101)

    assert np.all(np.isclose(emit_py.temperature.value, temp_idl))
    assert np.all(np.isclose(emit_py.probability, prob_idl))
    assert np.all(np.isclose(emit_py.probgrid, vgrid_idl))

    # import matplotlib.pyplot as plt
    # plt.plot(emit_py.probability, emit_py.probgrid[0,:])
    # plt.plot(prob_idl, vgrid_idl[0,:])
    # plt.show()


import numpy as np
from ..LossInfo import LossInfo


def test_LossInfo():
    rate0 = LossInfo('Na', -2., 0.3)
    assert rate0.photo == 0.5
    
    rate1 = LossInfo('Ca', 0, 1)
    assert rate1.photo == 7e-5
    
    rate2 = LossInfo('Mg', 0, 0.3)
    assert np.isclose(rate2.photo, 7.21111111111e-6)


from scipy.io import readsav


class IDLout:
    def __init__(self, filename):
        idl = readsav(filename)
        idlout = idl['output']

        self.x = idlout['x'][0]
        self.y = idlout['y'][0]
        self.z = idlout['z'][0]
        self.vx = idlout['vx'][0]
        self.vy = idlout['vy'][0]
        self.vz = idlout['vz'][0]
        self.frac = idlout['frac'][0]
        self.time = idlout['time'][0]
        self.index = idlout['index'][0]

        index = sorted(list(set(idlout['index'][0])))
        x0 = [idlout['x0'][0][idlout['index'][0] == i][0] for i in index]
        y0 = [idlout['z0'][0][idlout['index'][0] == i][0] for i in index]
        z0 = [idlout['y0'][0][idlout['index'][0] == i][0] for i in index]
        vx0 = [idlout['vx0'][0][idlout['index'][0] == i][0] for i in index]
        vy0 = [idlout['vy0'][0][idlout['index'][0] == i][0] for i in index]
        vz0 = [idlout['vz0'][0][idlout['index'][0] == i][0] for i in index]

        self.x0 = np.array(x0)
        self.y0 = np.array(y0)
        self.z0 = np.array(z0)
        self.vx0 = np.array(vx0)
        self.vy0 = np.array(vy0)
        self.vz0 = np.array(vz0)
        del idlout



"""Read the model inputs from a file and create the Input object.

The Input object is build from smaller objects defining different model
options.

Geometry
    Defines the Solar System geometry for the Input.

SurfaceInteraction
    Defines the surface interactions.

Forces
    Set which forces act on model particles.

SpatialDist
    Define the initial spatial distribution of particles.

SpeedDist
    Define the initial speed distribution of particles.

AngularDist
    Define the initial angular distribtuion of particles.

Options
    Configure other model parameters
"""
import os
import os.path
import numpy as np
import sys
import pandas as pd
import logging
from astropy.time import Time
import astropy.units as u
from .Output import Output
from .configure_model import configfile
from .database_connect import database_connect
from .input_classes import (Geometry, SurfaceInteraction, Forces, SpatialDist,
                            SpeedDist, AngularDist, Options)
from .produce_image import ModelImage


class Input:
    def __init__(self, infile):
        """Read the input options from a file.

        **Parameters**
        
        infile
            Plain text file containing model input parameters. See
            :doc:`inputfiles` for a description of the input file format.

        **Class Attributes**

        * geometry
        
        * surface_interaction
        
        * forces
        
        * spatialdist
        
        * speeddist
        
        * angulardist
        
        * options
        
        """
        # Read the configuration file
        self._savepath = configfile()

        # Read in the input file:
        self._inputfile = infile
        params = []
        if os.path.isfile(infile):
            # Remove everything in the line after a comment character
            for line in open(infile, 'r'):
                if ';' in line:
                    line = line[:line.find(';')]
                elif '#' in line:
                    line = line[:line.find('#')]
                else:
                    pass
                    
                if line.count('=') == 1:
                    param_, val_ = line.split('=')
                    if param_.count('.') == 1:
                        sec_, par_ = param_.split('.')
                        params.append((sec_.casefold().strip(),
                                       par_.casefold().strip(),
                                       val_.casefold().strip()))
                    else:
                        pass
                else:
                    pass
        else:
            raise FileNotFoundError(infile)
            
        def extract_param(tag):
            return {b:c for (a,b,c) in params if a == tag}

        self.geometry = Geometry(extract_param('geometry'))
        self.surfaceinteraction = SurfaceInteraction(extract_param(
            'surfaceinteraction'))
        self.forces = Forces(extract_param('forces'))
        self.spatialdist = SpatialDist(extract_param('spatialdist'))
        self.speeddist = SpeedDist(extract_param('speeddist'))
        self.angulardist = AngularDist(extract_param('angulardist'))
        self.options = Options(extract_param('options'))
        
    def __repr__(self):
        return self.__str__()

    def __str__(self):
        result = (self.geometry.__str__() + '\n' +
                  self.surfaceinteraction.__str__() + '\n' +
                  self.forces.__str__() + '\n' +
                  self.spatialdist.__str__() + '\n' +
                  self.speeddist.__str__() + '\n' +
                  self.angulardist.__str__() + '\n' +
                  self.options.__str__())
        
        return result

    def search(self):
        """ Search the database for previous model runs with the same inputs.
        See :doc:`searchtolerances` for tolerances used in searches.
        
        **Parameters**
        
        No parameters.
        
        **Returns**
        
        * A list of filenames corresponding to the inputs.
        
        * Number of packets contained in those saved outputs.
        
        * Total modeled source rate.
        """
        geo_id = self.geometry.search()
        sint_id = self.surfaceinteraction.search()
        for_id = self.forces.search()
        spat_id = self.spatialdist.search()
        spd_id = self.speeddist.search()
        ang_id = self.angulardist.search()
        opt_id = self.options.search()
        
        if None in [geo_id, sint_id, for_id, spat_id, spd_id, ang_id, opt_id]:
            return [], 0., 0.
        else:
            query = f'''SELECT idnum, filename, npackets, totalsource
                        FROM outputfile
                        WHERE geo_type = '{self.geometry.type}' and
                              geo_id = {geo_id} and
                              sint_type = '{self.surfaceinteraction.sticktype}' and
                              sint_id = {sint_id} and
                              force_id = {for_id} and
                              spatdist_type = '{self.spatialdist.type}' and
                              spatdist_id = {spat_id} and
                              spddist_type = '{self.speeddist.type}' and
                              spddist_id = {spd_id} and
                              angdist_type = '{self.angulardist.type}' and
                              angdist_id = {ang_id} and
                              opt_id = {opt_id}'''
            with database_connect() as con:
                result = pd.read_sql(query, con)
            
            return (result.filename.to_list(), result.npackets.sum(),
                    result.totalsource.sum())

    def run(self, npackets, packs_per_it=None, overwrite=False, compress=True):
        """Run the nexoclom model with the current inputs.
        
        **Parameters**
        
        npackets
            Number of packets to simulate
        
        packs_per_it
            Maximum number of packets to run at one time. Default = 1e5 in
            constant step-size mode; 1e6 in adaptive step-size mode.
        
        overwrite
            Erase any files matching the current inputs that exist.
            Default = False
            
        compress
            Remove packets with frac=0 from the outputs to reduce file size.
            Default = True
            
        **Outputs**
        
        Nothing is returned, but model runs are saved and cataloged.
        """
        # Configure the logger
        # Note: The logfile name will be changed to match the outputfile name
        #       and stored in the Output object.
        # logger = logging.getLogger()
        # logger.setLevel(logging.INFO)
        # log_file_handler = logging.FileHandler('log.out', 'w')
        # logger.addHandler(log_file_handler)
        # out_handler = logging.StreamHandler(sys.stdout)
        # logger.addHandler(out_handler)
        # fmt = logging.Formatter('%(levelname)s: %(msg)s')
        # log_file_handler.setFormatter(fmt)
        # out_handler.setFormatter(fmt)

        t0_ = Time.now()
        # logger.info(f'Starting at {t0_}')
        print(f'Starting at {t0_}')

        if len(self.geometry.planet) != 1:
            # logger.error('Gravity and impact check not working for '
            #               'planets with moons.')
            sys.exit()
            
        # Determine how many packets have already been run
        if overwrite:
            self.delete_files()
            totalpackets = 0
        else:
            outputfiles, totalpackets, _ = self.search()
            # logger.info(f'Found {len(outputfiles)} files with {totalpackets} '
            #              'packets.')
            print(f'Found {len(outputfiles)} files with {totalpackets} '
                  'packets.')

        npackets = int(npackets)
        ntodo = npackets - totalpackets
        
        if ntodo > 0:
            if (packs_per_it is None) and (self.options.step_size == 0):
                packs_per_it = 1000000
            elif packs_per_it is None:
                packs_per_it = (1e8 * self.options.step_size /
                                self.options.endtime.value)
            else:
                pass
            packs_per_it = int(np.min([ntodo, packs_per_it]))
            
            # Determine how many iterations are needed
            nits = int(np.ceil(ntodo/packs_per_it))
            
            # logger.info('Running Model')
            # logger.info(f'Will complete {nits} iterations of {packs_per_it} '
            #              'packets.')
            print('Running Model')
            print(f'Will complete {nits} iterations of {packs_per_it} packets.')

            for _ in range(nits):
                tit0_ = Time.now()
             #    logger.info(f'Starting iteration #{_+1} of {nits}')
                print(f'Starting iteration #{_+1} of {nits}')

                # Create an output object
                Output(self, packs_per_it, compress=compress)
                       #logger=logger)
                # Just run and save the model when output is created
                # No reason to explicitly call run
                
                tit1_ = Time.now()
                # logger.info(f'Completed iteration #{_+1} in '
                #             f'{(tit1_ - tit0_).sec} seconds.')
                print(f'Completed iteration #{_+1} in '
                      f'{(tit1_ - tit0_).sec} seconds.')
        else:
            pass

        t2_ = Time.now()
        dt_ = (t2_-t0_).sec
        if dt_ < 60:
            dt_ = f'{dt_} sec'
        elif dt_ < 3600:
            dt_ = f'{dt_/60} min'
        else:
            dt_ = f'{dt_/3600} hr'
        # logger.info(f'Model run completed in {dt_} at {t2_}.')
        # out_handler.close()
        print(f'Model run completed in {dt_} at {t2_}.')

    def produce_image(self, format_, filenames=None, overwrite=False):
        return ModelImage(self, format_, filenames=filenames,
                          overwrite=overwrite)
    
    def delete_files(self):
        """Delete output files and remove them from the database.

        **Parameters**

        filelist
            List of files to remove. This can be found with Inputs.search()

        **Returns**

        No outputs.

        """
        filelist, _, _ = self.search()
        with database_connect() as con:
            cur = con.cursor()
            
            for f in filelist:
                # Delete the file
                print(f'Deleting {f}')
                if os.path.exists(f):
                    os.remove(f)
                
                # Remove from database
                cur.execute('''SELECT idnum FROM outputfile
                               WHERE filename = %s''', (f,))
                idnum = cur.fetchone()[0]
                
                cur.execute('''DELETE FROM outputfile
                               WHERE idnum = %s''', (idnum,))
                
                cur.execute('''SELECT idnum, filename FROM modelimages
                               WHERE out_idnum = %s''', (idnum,))
                for mid, mfile in cur.fetchall():
                    cur.execute('''DELETE from modelimages
                                   WHERE idnum = %s''', (mid,))
                    if os.path.exists(mfile):
                        os.remove(mfile)



import numpy as np
from scipy import interpolate
import astropy.units as u
import astropy.constants as const
from atomicdataMB import atomicmass
from .surface_temperature import surface_temperature
from .source_distribution import MaxwellianDist

def temp_dependent_sticking_coef(tsurf, A):
    return


class SurfaceInteraction:
    def __init__(self, inputs, **kwargs):
        # Sticking coeffienct setup
        if inputs.surfaceinteraction.sticktype == 'temperature dependent':
            A = inputs.surfaceinteraction.A
            def stickcoef(lon, lat):
                tsurf = surface_temperature(inputs.geometry, lon, lat)
                stickcoef = A[0] * np.exp(A[1]*tsurf) + A[2]
                stickcoef[stickcoef > 1.] = 1.
                stickcoef[stickcoef < 0.] = 0.
                return stickcoef
            
            self.stickcoef = stickcoef
        elif inputs.surfaceinteraction.sticktype == 'surface map':
            assert 0
        else:
            pass

        # Surface accommodation setup
        if inputs.surfaceinteraction.accomfactor == 0:
            pass
        else:
            longitude = np.arange(361)*np.pi/180.
            latitude = np.arange(181)*np.pi/180. - np.pi/2.
            longrid, latgrid = np.meshgrid(longitude, latitude)
            tsurf = surface_temperature(inputs.geometry, longrid.flatten(),
                                        latgrid.flatten())
            
            nt = kwargs['nt'] if 'nt' in kwargs else 201
            nv = kwargs['nv'] if 'nv' in kwargs else 101
            nprob = kwargs['nprob'] if 'nprob' in kwargs else 101
            
            temperature = np.linspace(min(tsurf), max(tsurf), nt)*u.K
            v_temp = np.sqrt(2*temperature*const.k_B/
                             atomicmass(inputs.options.species))
            v_temp = v_temp.to(u.km/u.s)
            probability = np.linspace(0, 1, nprob)
            probgrid = np.ndarray((nt,nprob))
            for i,t in enumerate(temperature):
                vrange = np.linspace(0*u.km/u.s, v_temp[i]*3, nv)
                f_v = MaxwellianDist(vrange, t, inputs.options.species)
                cumdist = f_v.cumsum()
                cumdist -= cumdist.min()
                cumdist /= cumdist.max()
                probgrid[i,:] = np.interp(probability, cumdist, vrange.value)
                
            self.v_interp = interpolate.RectBivariateSpline(temperature.value,
                                                            probability,
                                                            probgrid).ev
            self.probgrid = probgrid
            self.temperature = temperature
            self.probability = probability
            



import pickle
import numpy as np
from scipy.spatial import distance, distance_matrix
import timeit


def oldway(packets, xdata, boresight):
    # Old way
    rpr = np.ndarray((len(packets), len(xdata)))
    for i in range(len(xdata)):
        row_pr = xdata[i, :]
        xpr = packets-row_pr[np.newaxis, :]
        #bore = boresight[i, :]
        rpr[:,i] = np.linalg.norm(xpr, axis=1)
        
        # Packet-s/c boresight angle
        #losrad = np.sum(xpr*bore[np.newaxis, :], axis=1)
        # costheta = losrad/rpr
        # costheta[costheta > 1] = 1.
        # costheta[costheta < -1] = -1.
        
    return rpr

def newway(packets, xdata, boresight):
    for i in range(len(xdata)):
        row_pr = xdata[i, :]
        xpr = packets-row_pr[np.newaxis, :]
        bore = boresight[i, :]
        losrad = distance.cdist(xpr, bore[np.newaxis, :],
                                metric=lambda u, v: np.dot(u, v)).flatten()
    
    return losrad

def thirdway(packets, xdata, boresight):
    rpr = distance_matrix(packets, xdata)
    return rpr

packets, xdata, boresight = pickle.load(open('packets.pkl', 'rb'))
#packets = packets[0:1000,:]
#xdata = xdata[0:10,:]
#boresight = boresight[0:10,:]

#print(timeit.timeit('oldway(packets[:,2:5], xdata, boresight)', globals=globals()))
#print(timeit.timeit('newway(packets[:,2:5], xdata, boresight)', globals=globals()))
#print(timeit.timeit('thirdway(packets[:,2:5], xdata, boresight)', globals=globals()))

from datetime import datetime

t0 = datetime.now()
rpr0 = oldway(packets[:,2:5], xdata, boresight)
t1 = datetime.now()
rpr1 = thirdway(packets[:,2:5], xdata, boresight)
t2 = datetime.now()

print(t1-t0, t2-t1)




def satellite_initial_positions(inputs):
    '''
    Calculates the initial x,y,z of each satellite.
    If satellite motion is included, then it gives the locations of the
    satellites at each time during the simulation.

    Going to wait until I have some satellites to deal with to write this.
    '''

    time_given = 'time' in inputs.geometry

    return None



"""Test fails because of issues at the terminator and south pole.
I'm pretty sure the error is on the IDL side, but the effect is
negligible."""

import numpy as np
import os.path
from scipy.io import readsav
try:
    from .surface_temperature import surface_temperature
    from .Input import Input
except:
    from nexoclom.surface_temperature import surface_temperature
    from nexoclom import Input

idlout = readsav(os.path.join(os.path.dirname(__file__),
                              'data', 'tsurf.sav'))
t_idl, cos_idl = idlout['tsurf'], idlout['cosgrid']
lon_idl, lat_idl = idlout['longrid'], idlout['latgrid']

inputs = Input(os.path.join(os.path.dirname(__file__), 'inputfiles',
                            'Na.maxwellian.1200.accom.input'))
longitude = np.arange(361)*np.pi/180.
latitude = np.arange(181)*np.pi/180.-np.pi/2.
longrid, latgrid = np.meshgrid(longitude, latitude)
cosgrid = np.abs(np.cos(longrid) * np.cos(latgrid))**.25
tsurf = surface_temperature(inputs.geometry, longrid.flatten(),
                            latgrid.flatten()).reshape(181, 361)

#x, y = np.where(np.logical_not(np.isclose(tsurf, t_idl)))

assert np.all(np.isclose(tsurf[1:-1,:], t_idl[1:-1,:]))


import os.path
import astropy.units as u
import numpy as np
from scipy.io import readsav
import pandas as pd
import pickle
import subprocess
from astropy.visualization import PercentileInterval
try:
    from ..Input import Input
    from ..Output import Output
except:
    from nexoclom import Input, Output
from MESSENGERuvvs import MESSENGERdata
from bokeh.layouts import column
import bokeh.plotting as bkp


inputfiles = ['Ca.isotropic.maxwellian.50000.input',    # Good
              'Ca.spot.maxwellian.input',               # Good
              'Na.Gaussian.4_2.south.input',            # Good
              'Na.Sputtering.input',                    # Good
              'Ca.Gaussian.4_2.south.input',            # Good
              'Ca.spot.gaussian.input',                 # Good
              'Na.Gaussian.3_1.input',                  # Good
              'Na.Gaussian.3_1.no_accom.input',         # Good
              'Na.maxwellian.1200.accom.input',         # fixed
              'Na.Gaussian.3_1.accom.input',
              'Na.Gaussian.3_1.fullaccom.input']

idloverwrite = False
pyoverwrite = False

def load_idl_files(filename):
    X0 = {'x0':[], 'y0':[], 'z0':[], 'vx0':[], 'vy0':[], 'vz0':[]}
    X = {'x':[], 'y':[], 'z':[], 'vx':[], 'vy':[], 'vz':[], 'frac':[]}
    idlout = readsav(filename)
    
    for line in idlout['outputfile']:
        out = readsav(line.strip())
        out = out['output']
        t0 = out['time'][0] == max(out['time'][0])
        for key in X0.keys():
            X0[key].extend(out[key][0][t0])
        for key in X.keys():
            X[key].extend(out[key][0])
            
    result = pd.DataFrame(X0)
    result2 = pd.DataFrame(X)
    
    return (result, result2, idlout['im'], idlout['col'], idlout['radiance'],
            idlout['packets'])

for infile_ in inputfiles:
    infile = os.path.join(os.path.dirname(__file__),
                          'inputfiles', infile_)
    
    idlsavefile = os.path.join(os.path.dirname(__file__), 'outputs',
                               infile_.replace('.input', '.IDL.pkl'))
    idlfile = f"'{infile.replace('.input', '.IDL.input')}'"
    idlfile_ = infile.replace('.input', '.IDL.input')
    idloutputfile = os.path.join(os.path.dirname(__file__), 'idloutputs',
                                 os.path.basename(idlfile_)+'.sav')
    if not(os.path.exists(idlsavefile)) or idloverwrite:
        # Run the IDL version
        if not(idloutputfile) or idloverwrite:
            print('Running IDL version')
            ow = 1 if idloverwrite else 0
            command = (f'/Applications/exelis/idl85/bin/idl '
                       f'-e "run_idl_version, {idlfile}, overwrite={ow}"')
            print(command)
            idlcmd = subprocess.run(command, shell=True, capture_output=True)
            assert 'Completed Successfully' in str(idlcmd.stdout), idlcmd
        else:
            pass

        # Load the IDL results
        print(f'Loading from {idloutputfile}')
        idl0, idl, idlim, idlcol, idlrad, idlpack = load_idl_files(idloutputfile)
        idlim = idlim.transpose()
        idlcol = idlcol.transpose()

        with open(idlsavefile, 'wb') as f:
            pickle.dump((idl0, idl, idlim, idlcol, idlrad, idlpack), f)
    else:
        print('Found IDL results')
        with open(idlsavefile, 'rb') as f:
            idl0, idl, idlim, idlcol, idlrad, idlpack = pickle.load(f)
    
    # Run the python version
    inputs = Input(infile)
    sp = inputs.options.species
    orbit = 1489 if sp == 'Ca' else 1541
    data = MESSENGERdata(sp, f'orbit = {orbit}')
    inputs.geometry.taa = data.taa * u.rad
    data.model(inputs, 1e5, overwrite=pyoverwrite)
    pyim = inputs.produce_image('inputfiles/MercuryEmission.format')
    pycol = inputs.produce_image('inputfiles/MercuryColumn.format')

    # Load the python results
    outfiles, _, _ = inputs.search()
    output = Output.restore(outfiles[0])
    py0 = output.X0
    py = output.X

    # Convert velocities to km/s
    py0.vx = py0.vx * output.inputs.geometry.planet.radius
    py0.vy = py0.vy * output.inputs.geometry.planet.radius
    py0.vz = py0.vz * output.inputs.geometry.planet.radius
    py.vx = py.vx * output.inputs.geometry.planet.radius
    py.vy = py.vy * output.inputs.geometry.planet.radius
    py.vz = py.vz * output.inputs.geometry.planet.radius

    idl0.vx0 = idl0.vx0 * output.inputs.geometry.planet.radius
    idl0.vy0 = idl0.vy0 * output.inputs.geometry.planet.radius
    idl0.vz0 = idl0.vz0 * output.inputs.geometry.planet.radius
    idl.vx = idl.vx * output.inputs.geometry.planet.radius
    idl.vy = idl.vy * output.inputs.geometry.planet.radius
    idl.vz = idl.vz * output.inputs.geometry.planet.radius

    # Compare initial spatial distribution
    xbins = np.linspace(-1, 1, 21)
    xpy0, _ = np.histogram(py0.x, bins=xbins)
    ypy0, _ = np.histogram(py0.y, bins=xbins)
    zpy0, _ = np.histogram(py0.z, bins=xbins)

    xidl0, _ = np.histogram(idl0.x0, bins=xbins)
    yidl0, _ = np.histogram(idl0.y0, bins=xbins)
    zidl0, _ = np.histogram(idl0.z0, bins=xbins)

    p1 = bkp.figure(plot_width=1000, plot_height=600,
                    title=f'{infile_}: Initial x Distribution')
    #p1.legend.click_policy = 'hide'
    p1.line(xbins[0:-1], xpy0/max(xpy0), color='red', legend_label='Python')
    p1.line(xbins[0:-1], xidl0/max(xidl0), color='blue', legend_label='IDL')

    p2 = bkp.figure(plot_width=1000, plot_height=600,
                    title = f'{infile_}: Initial y Distribution')
    p2.line(xbins[0:-1], ypy0/max(ypy0), color='red', legend_label='Python')
    p2.line(xbins[0:-1], yidl0/max(yidl0), color='blue', legend_label='IDL')

    p3 = bkp.figure(plot_width=1000, plot_height=600,
                    title = f'{infile_}: Initial z Distribution')
    p3.line(xbins[0:-1], zpy0/max(zpy0), color='red', legend_label='Python')
    p3.line(xbins[0:-1], zidl0/max(zidl0), color='blue', legend_label='IDL')
    
    # compare initial velocity distribution
    vbins = np.linspace(np.min(py0.vx), np.max(py0.vx), 21)
    vxpy0, _ = np.histogram(py0.vx, bins=vbins)
    vypy0, _ = np.histogram(py0.vy, bins=vbins)
    vzpy0, _ = np.histogram(py0.vz, bins=vbins)

    vxidl0, _ = np.histogram(idl0.vx0, bins=vbins)
    vyidl0, _ = np.histogram(idl0.vy0, bins=vbins)
    vzidl0, _ = np.histogram(idl0.vz0, bins=vbins)

    p4 = bkp.figure(plot_width=1000, plot_height=600,
                    title=f'{infile_}: Initial vx Distribution')
    p4.line(xbins[0:-1], vxpy0/max(vxpy0), color='red', legend_label='Python')
    p4.line(xbins[0:-1], vxidl0/max(vxidl0), color='blue', legend_label='IDL')

    p5 = bkp.figure(plot_width=1000, plot_height=600,
                    title = f'{infile_}: Initial vy Distribution')
    p5.line(xbins[0:-1], vypy0/max(vypy0), color='red', legend_label='Python')
    p5.line(xbins[0:-1], vyidl0/max(vyidl0), color='blue', legend_label='IDL')

    p6 = bkp.figure(plot_width=1000, plot_height=600,
                    title = f'{infile_}: Initial vz Distribution')
    p6.line(xbins[0:-1], vzpy0/max(vzpy0), color='red', legend_label='Python')
    p6.line(xbins[0:-1], vzidl0/max(vzidl0), color='blue', legend_label='IDL')

    # Compare final spatial distribution
    xbins = np.linspace(-10, 10, 201)
    xpy, _ = np.histogram(py.x, bins=xbins)
    ypy, _ = np.histogram(py.y, bins=xbins)
    zpy, _ = np.histogram(py.z, bins=xbins)

    xidl, _ = np.histogram(idl.x, bins=xbins)
    yidl, _ = np.histogram(idl.y, bins=xbins)
    zidl, _ = np.histogram(idl.z, bins=xbins)

    p7 = bkp.figure(plot_width=1000, plot_height=600,
                    title=f'{infile_}: Final x Distribution')
    p7.line(xbins[0:-1], xpy/max(xpy), color='red', legend_label='Python')
    p7.line(xbins[0:-1], xidl/max(xidl), color='blue', legend_label='IDL')

    p8 = bkp.figure(plot_width=1000, plot_height=600,
                    title=f'{infile_}: Final y Distribution')
    p8.line(xbins[0:-1], ypy/max(ypy), color='red', legend_label='Python')
    p8.line(xbins[0:-1], yidl/max(yidl), color='blue', legend_label='IDL')

    p9 = bkp.figure(plot_width=1000, plot_height=600,
                    title=f'{infile_}: Final z Distribution')
    p9.line(xbins[0:-1], zpy/max(zpy), color='red', legend_label='Python')
    p9.line(xbins[0:-1], zidl/max(zidl), color='blue', legend_label='IDL')

    # Compare final velocity distribution
    vbins = np.linspace(np.min(py.vx), np.max(py.vx), 201)
    vxpy, _ = np.histogram(py.vx, bins=vbins, weights=py.frac)
    vypy, _ = np.histogram(py.vy, bins=vbins, weights=py.frac)
    vzpy, _ = np.histogram(py.vz, bins=vbins, weights=py.frac)

    vxidl, _ = np.histogram(idl.vx, bins=vbins, weights=idl.frac)
    vyidl, _ = np.histogram(idl.vy, bins=vbins, weights=idl.frac)
    vzidl, _ = np.histogram(idl.vz, bins=vbins, weights=idl.frac)

    p10 = bkp.figure(plot_width=1000, plot_height=600,
                    title=f'{infile_}: Final vx Distribution')
    p10.line(vbins[0:-1], vxpy/max(vxpy), color='red', legend_label='Python')
    p10.line(vbins[0:-1], vxidl/max(vxidl), color='blue', legend_label='IDL')

    p11 = bkp.figure(plot_width=1000, plot_height=600,
                    title=f'{infile_}: Final vy Distribution')
    p11.line(vbins[0:-1], vypy/max(vypy), color='red', legend_label='Python')
    p11.line(vbins[0:-1], vyidl/max(vyidl), color='blue', legend_label='IDL')

    p12 = bkp.figure(plot_width=1000, plot_height=600,
                    title=f'{infile_}: Final vz Distribution')
    p12.line(vbins[0:-1], vzpy/max(vzpy), color='red', legend_label='Python')
    p12.line(vbins[0:-1], vzidl/max(vzidl), color='blue', legend_label='IDL')

    ###############################################

    q0 = np.sum(pyim.image[:,200:250], axis=1)
    q1 = np.sum(pyim.image[:,250:300], axis=1)
    q2 = np.sum(pyim.image[200:250,:], axis=0)
    q3 = np.sum(pyim.image[250:300,:], axis=0)
    i0 = np.sum(idlim[:,200:250], axis=1)
    i1 = np.sum(idlim[:,250:300], axis=1)
    i2 = np.sum(idlim[200:250,:], axis=0)
    i3 = np.sum(idlim[250:300,:], axis=0)
    
    p13 = bkp.figure(plot_width=1000, plot_height=600,
                    title=f'{infile_}: Radiance')
    p13.line(pyim.xaxis, q0, legend_label='py0', color='red')
    p13.line(pyim.xaxis, i0, legend_label='idl0', color='blue')
    p13.line(pyim.xaxis, q1, legend_label='py1', color='red')
    p13.line(pyim.xaxis, i1, legend_label='idl1', color='blue')

    p14 = bkp.figure(plot_width=1000, plot_height=600,
                    title=f'{infile_}: Radiance')
    p14.line(pyim.xaxis, q2, legend_label='py2', color='red')
    p14.line(pyim.xaxis, i2, legend_label='idl2', color='blue')
    p14.line(pyim.xaxis, q3, legend_label='py3', color='red')
    p14.line(pyim.xaxis, i3, legend_label='idl3', color='blue')
    
    q0 = np.sum(pycol.image[:, 200:250], axis=1)
    q1 = np.sum(pycol.image[:, 250:300], axis=1)
    q2 = np.sum(pycol.image[200:250, :], axis=0)
    q3 = np.sum(pycol.image[250:300, :], axis=0)
    i0 = np.sum(idlcol[:, 200:250], axis=1)
    i1 = np.sum(idlcol[:, 250:300], axis=1)
    i2 = np.sum(idlcol[200:250, :], axis=0)
    i3 = np.sum(idlcol[250:300, :], axis=0)

    p15 = bkp.figure(plot_width=1000, plot_height=600,
                    title=f'{infile_}: Column')
    p15.line(pycol.xaxis, q0, legend_label='py0', color='red')
    p15.line(pycol.xaxis, i0, legend_label='idl0', color='blue')
    p15.line(pycol.xaxis, q1, legend_label='py1', color = 'red')
    p15.line(pycol.xaxis, i1, legend_label='idl1', color='blue')

    p16 = bkp.figure(plot_width=1000, plot_height=600,
                    title=f'{infile_}: Column')
    p16.line(pycol.xaxis, q2, legend_label='py2', color = 'red')
    p16.line(pycol.xaxis, i2, legend_label='idl2', color='blue')
    p16.line(pycol.xaxis, q3, legend_label='py3', color = 'red')
    p16.line(pycol.xaxis, i3, legend_label='idl3', color='blue')

    # Estimate model strength (source rate) by mean of middle 50%
    interval = PercentileInterval(50)
    lim = interval.get_limits(data.data.radiance)
    mask = ((data.data.radiance >= lim[0]) &
            (data.data.radiance <= lim[1]))
    m_data = np.mean(data.data.radiance[mask])
    m_idl = np.mean(idlrad[mask])
    str = m_data/m_idl

    p17 = bkp.figure(plot_width=1000, plot_height=600,
                     title=f'{infile_}: Orbit compare radiance')
    p17.circle(data.data['utc'], data.data['radiance'], color='black',
             legend_label='Data')
    p17.line(data.data['utc'], data.data['model0'], color='red',
             legend_label='Python')
    p17.line(data.data['utc'], idlrad*str, color='blue', legend_label='IDL')
    
    p18 = bkp.figure(plot_width=1000, plot_height=600,
                     title=f'{infile_}: Orbit compare packets')
    p18.line(data.data['utc'], data.data['packets0'], color='red',
             legend_label='Python')
    p18.line(data.data['utc'],
             idlpack*np.max(data.data['packets0'])/np.max(idlpack),
             color='blue', legend_label='IDL')

    savefile = os.path.join(os.path.dirname(__file__), 'figures',
                            infile_.replace('.input', '.html'))
    bkp.output_file(savefile)
    bkp.save(column(p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11,
                    p12, p13, p14, p15, p16, p17, p18))
 



import numpy as np
from scipy import interpolate
import astropy.units as u
import astropy.constants as const
from atomicdataMB import atomicmass
from .surface_temperature import surface_temperature
from .source_distribution import MaxwellianDist

def surface_interaction_setup(inputs):
    # Set up accommodation factor
    if inputs.surfaceinteraction.accomfactor != 0:
        longitude = np.arange(361)*np.pi/180.
        latitude = np.arange(181)*np.pi/180. - np.pi/2.
        longrid, latgrid = np.meshgrid(longitude, latitude)
        tsurf = surface_temperature(inputs.geometry, longrid.flatten(),
                                    latgrid.flatten())

        nt, nv, nprob = 201, 101, 101
        temperature = np.linspace(min(tsurf), max(tsurf), nt)*u.K
        v_temp = np.sqrt(2*temperature*const.k_B/
                         atomicmass(inputs.options.species))
        v_temp = v_temp.to(u.km/u.s)
        probability = np.linspace(0, 1, nprob)
        probgrid = np.ndarray((nt,nprob))
        for i,t in enumerate(temperature):
            vrange = np.linspace(0*u.km/u.s, np.max(v_temp[i]*3), nv)
            f_v = MaxwellianDist(vrange, t, inputs.options.species)
            cumdist = f_v.cumsum()
            cumdist -= cumdist.min()
            cumdist /= cumdist.max()
            probgrid[i,:] = np.interp(vrange.value, cumdist, probability)
            
        # v_interp = interpolate.interp2d(probability, temperature,
        #                                 probgrid)
        v_interp = interpolate.RectBivariateSpline(temperature.value,
                                                   probability, probgrid)
    else:
        v_interp = None
        
    surfaceint = {'v_accom': v_interp}
    
    return surfaceint
    


"""Test that all types of inputfiles can be read successfully.

To Test:
    (g.1) Geometry With Time Stamp
    (g.2) Geometry Without Time Stamp
    (g.3) Startpoint given
    (g.4) Startpoint not given
    (g.5) Specifiy objects for planet with moon
    (g.6) Don't specify objects for planet with moon
    (g.7) phi for planets with moon
    
    (si.1) Complete sticking
    (si.2) Constant sticking with some bouncing
    (si.3) Temperature dependent sticking
    (si.4) Sticking coefficient from a surface map
    
    (f.1) Gravity on
    (f.2) Gravity off
    (f.3) Radiation pressure on
    (f.4) Radiation pressure off
    
"""
import os.path
try:
    from ..Input import Input
except:
    from nexoclom import Input
from MESSENGERuvvs import MESSENGERdata

inputfiles = [#'Na.Gaussian.3_1.no_accom_tempsticking.input']
              #'Na.Gaussian.3_1.no_accom.input',
              # 'Na.maxwellian.1200.accom.input',
              #'Ca.surfacemap.maxwellian.50000.input',
              'Ca.spot.maxwellian.input']

orbit = 36
overwrite = False

def test_Input():
    for infile in inputfiles:
        data = MESSENGERdata(infile[0:2], f'orbit={orbit}')
        inputs = Input(os.path.join(os.path.dirname(__file__),
                                    'inputfiles', infile))
        # inputs.run(1e5, overwrite=overwrite)
        # image = inputs.produce_image('inputfiles/MercuryEmission.format',
        #                              overwrite=overwrite)
        # sfile = os.path.join(os.path.dirname(__file__),
        #                      'outputs', infile.replace('.input', '.html'))
        # image.display(show=True, savefile=sfile)

        data.model(inputs, 1e5, overwrite=overwrite, fit_method='middle50',
                   label='Middle 50')
        data.model(inputs, 1e5, overwrite=overwrite, fit_method='middle10',
                   label='Middle 10')
        data.model(inputs, 1e5, overwrite=overwrite, fit_method='middle90',
                   label='Middle 90')

        sfile = os.path.join(os.path.dirname(__file__), 'figures',
                             f'{infile}_Orbit{orbit:04d}.html')
        data.plot(filename=sfile)


    assert True
    return data

if __name__ == '__main__':
    data = test_Input()



'''
Computes the magnetic coordinates of each packet

Also determines whether the packet is in shadow
'''

import numpy as np

def xyz_to_magcoord(t, x, inputs, planet):
    if planet.object == 'Mercury':
        magcoord = None
    else:
        # Need to add magnetic coordinates for Jupiter and Saturn
        assert 0, '{} not impletmented'.format(planet)

    return magcoord



from nexoclom import Input
from MESSENGERuvvs import MESSENGERdata


data = MESSENGERdata('Ca', 'Orbit = 36')
data = data[0:-1]
inputs = Input('inputfiles/Ca.isotropic.maxwellian.50000.input')

# los0 = inputs.line_of_sight(data.data, 'radiance', version='test',
#                            overwrite=True)

los0 = inputs.line_of_sight(data.data, 'radiance', version='new',
                            overwrite=True)
# los1 = inputs.line_of_sight(data.data, 'radiance', version='old',
#                             overwrite=True)
import pickle
los1 = pickle.load(open('los1.pkl', 'rb'))

import matplotlib.pyplot as plt
plt.plot(los0.radiance)
plt.plot(los1.radiance, linestyle='--')
plt.show()
assert 0




import os.path
import numpy as np
import numpy.matlib as npmat
import pandas as pd
import pickle
import astropy.units as u
from solarsystemMB import SSObject
from mathMB import rotation_matrix
from .ModelResults import ModelResult
from .database_connect import database_connect
from .input_classes import InputError
from .Output import Output


class ModelImage(ModelResult):
    def __init__(self, inputs, format, filenames=None, overwrite=False):
        """ Create Images from model results.
        This Assumes the model has already been run.
        
        Parameters
        ==========
        inputs
            An Input object
        
        format
            A dictionary with format information or a path to a formatfile.
            
        filenames
            A filename or list of filenames to use. Default = None is to
            find all files created for the inputs.
            
        overwrite
            If True, deletes any images that have already been computed.
            Default = False
        """
        super().__init__(inputs, format, filenames=filenames)
        self.type = 'image'
        
        if 'origin' in self.format:
            self.origin = SSObject(self.format['origin'])
        else:
            self.origin = inputs.geometry.planet
            
        self.unit = u.def_unit('R_' + self.origin.object,
                               self.origin.radius)

        if 'dims' in self.format:
            dimtemp = self.format['dims'].split(',')
            self.dims = [int(dimtemp[0]), int(dimtemp[1])]
        else:
            self.dims = [800, 800]

        if 'center' in self.format:
            centtemp = self.format['center'].split(',')
            self.center = [float(centtemp[0])*self.unit,
                           float(centtemp[1])*self.unit]
        else:
            self.center = [0*self.unit, 0*self.unit]

        if 'width' in self.format:
            widtemp = self.format['width'].split(',')
            self.width = [float(widtemp[0])*self.unit,
                          float(widtemp[1])*self.unit]
        else:
            self.width = [8*self.unit, 8*self.unit]

        if 'subobslongitude' in self.format:
            self.subobslongitude = float(self.format['subobslongitude'])*u.rad
        else:
            self.subobslongitude = 0.*u.rad

        if 'subobslongitude' in self.format:
            self.subobslatitude = float(self.format['subobslatitude'])*u.rad
        else:
            self.subobslatitude = np.pi*u.rad

        self.image = np.zeros(self.dims)
        self.packet_image = np.zeros(self.dims)
        self.blimits = None
        immin = tuple(c - w/2 for c, w in zip(self.center, self.width))
        immax = tuple(c + w/2 for c, w in zip(self.center, self.width))
        scale = tuple(w/(d-1) for w, d in zip(self.width, self.dims))
        self.Apix = (scale[0]*scale[1]).to(u.cm**2)

        self.xaxis = np.linspace(immin[0], immax[0], self.dims[0])
        self.zaxis = np.linspace(immin[1], immax[1], self.dims[1])

        for i, fname in enumerate(self.filenames):
            # Search to see if its already been done
            print(f'Output filename: {fname}')
            image_, packets_ = self.restore(fname, overwrite=overwrite)

            if image_ is None:
                image_, packets_ = self.create_image(fname)
                print(f'Completed image {i+1} of {len(self.filenames)}')
            else:
                print(f'Image {i+1} of {len(self.filenames)} '
                       'previously completed.')

            self.image += image_
            self.packet_image += packets_

        self.image *= self.atoms_per_packet

    def save(self, fname, image, packets):
        # Determine the id of the outputfile
        idnum = int(os.path.basename(fname).split('.')[0])

        # Insert the image into the database
        if self.quantity == 'radiance':
            mech = ', '.join(sorted([m for m in self.mechanism]))
            wave_ = sorted([w.value for w in self.wavelength])
            wave = ', '.join([str(w) for w in wave_])
        else:
            mech = None
            wave = None

        width = [w.value for w in self.width]
        center = [c.value for c in self.center]
        
        with database_connect() as con:
            cur = con.cursor()
            cur.execute(f'''INSERT into modelimages (out_idnum, quantity,
                                origin, dims, center, width, subobslongitude,
                                subobslatitude, mechanism, wavelength,
                                filename)
                            VALUES (%s, %s, %s, %s::INT[2],
                                    %s::DOUBLE PRECISION[2],
                                    %s::DOUBLE PRECISION[2],
                                    %s, %s, %s, %s, 'temp')''',
                        (idnum, self.quantity, self.origin.object,
                         self.dims, center, width,
                         self.subobslongitude.value, self.subobslatitude.value,
                         mech, wave))

        idnum_ = pd.read_sql(f'''SELECT idnum
                                FROM modelimages
                                WHERE filename = 'temp';''', con)
        assert len(idnum_) == 1
        idnum = int(idnum_.idnum[0])

        savefile = os.path.join(os.path.dirname(fname), f'image.{idnum}.pkl')
        with open(savefile, 'wb') as f:
            pickle.dump((image, packets), f)
        cur.execute(f'''UPDATE modelimages
                        SET filename=%s
                        WHERE idnum = %s''', (savefile, idnum))
        con.close()

    def restore(self, fname, overwrite=False):
        # Determine the id of the outputfile
        if self.quantity == 'radiance':
            mech = ("mechanism = '" +
                    ", ".join(sorted([m for m in self.mechanism])) +
                    "'")
            wave_ = sorted([w.value for w in self.wavelength])
            wave = ("wavelength = '" +
                    ", ".join([str(w) for w in wave_]) +
                    "'")
        else:
            mech = 'mechanism is NULL'
            wave = 'wavelength is NULL'

        with database_connect() as con:
            idnum_ = pd.read_sql(f'''SELECT idnum
                                FROM outputfile
                                WHERE filename='{fname}' ''', con)
            oid = idnum_.idnum[0]

            result = pd.read_sql(
                f'''SELECT filename FROM modelimages
                    WHERE out_idnum = {oid} and
                          quantity = '{self.quantity}' and
                          origin = '{self.origin.object}' and
                          dims[1] = {self.dims[0]} and
                          dims[2] = {self.dims[1]} and
                          center[1] = {self.center[0].value} and
                          center[2] = {self.center[1].value} and
                          width[1] = {self.width[0].value} and
                          width[2] = {self.width[1].value} and
                          subobslongitude = {self.subobslongitude.value} and
                          subobslatitude = {self.subobslatitude.value} and
                          {mech} and
                          {wave}''', con)

        if (len(result) == 1) and overwrite:
            if os.path.exists(result.filename[0]):
                os.remove(result.filename[0])
            with database_connect() as con:
                cur = con.cursor()
                cur.execute('''DELETE FROM modelimages
                               WHERE filename = %s''', (result.filename[0],))
            image, packets = None, None
        elif len(result) == 1:
            image, packets = pickle.load(open(result.filename[0], 'rb'))
        elif len(result) == 0:
            image, packets = None, None
        else:
            raise RuntimeError('ModelImage.restore',
                               'Should not be able to get here.')

        return image, packets

    def create_image(self, fname):
        # Determine the proper frame rotation
        M = self.image_rotation()

        # Load data in solar reference frame
        output = Output.restore(fname)
        packets = output.X
        if self.origin != self.inputs.geometry.planet:
            super().transform_reference_frame(packets)
        
        packets['radvel_sun'] = (packets['vy'] +
                                 output.vrplanet.to(self.unit/u.s).value)

        # packet positions in an array
        pts_sun = packets[['x', 'y', 'z']].values

        # Rotate to observer frame
        pts_obs = np.array(np.matmul(M, pts_sun.transpose()).transpose())

        # Determine which packets are not blocked by planet
        rhosqr_obs = np.linalg.norm(pts_obs[:, [0, 2]], axis=1)
        inview = (rhosqr_obs > 1) | (pts_obs[:,1] < 0)
        packets['frac'] *= inview

        # Which packets are in sunlight
        rhosqr_sun = np.linalg.norm(pts_sun[:, [0, 2]], axis=1)
        out_of_shadow = (rhosqr_sun > 1) | (pts_sun[:,1] < 0)

        # Packet weighting
        self.packet_weighting(packets, out_of_shadow, output.aplanet)
        packets['weight'] /= self.Apix

        # Edges of each pixel (bin) for the histogram
        dx = (self.xaxis[1]-self.xaxis[0])/2.
        bx = np.append(self.xaxis-dx, self.xaxis[-1]+dx)
        dz = (self.zaxis[1]-self.zaxis[0])/2.
        bz = np.append(self.zaxis-dz, self.zaxis[-1]+dz)

        pts_obs = pts_obs.transpose()
        image, _, _ = np.histogram2d(pts_obs[0,:], pts_obs[2,:],
                                     weights=packets['weight'], bins=(bx, bz))
        packim, _, _ = np.histogram2d(pts_obs[0,:], pts_obs[2,:],
                                       bins=(bx, bz))

        self.save(fname, image, packim)
        del output

        return image, packim

    def display(self, savefile='image.png', limits=None, show=True):
        import bokeh.plotting as bkp
        from bokeh.palettes import Inferno256
        from bokeh.models import (HoverTool, ColumnDataSource, ColorBar,
                                  LogColorMapper, LogTicker)
        from bokeh.io import curdoc
        from bokeh.themes import Theme

        if self.unit.__str__() == 'R_Mercury':
            ustr = 'R_M'
        else:
            ustr = 'R_obj'
            
        if self.quantity == 'radiance':
            runit = 'kR'
            rname = 'Radiance'
        elif self.quantity == 'column':
            runit = 'cm-2'
            rname = 'Column'
        else:
            assert 0

        tooltips = [('x', '$x{0.1f} ' + ustr),
                    ('y', '$y{0.1f} ' + ustr),
                    (rname, '@image ' + runit)]
        
        curdoc().theme = Theme(os.path.join(os.path.dirname(__file__),
                                            'data', 'bokeh.yml'))

        low = np.min(self.image[self.image > 0])
        color_mapper = LogColorMapper(palette=Inferno256, low=low,
                                      high=np.max(self.image))

        x0 = np.min(self.xaxis.value)
        y0 = np.min(self.zaxis.value)
        dw = np.max(self.xaxis.value) - np.min(self.xaxis.value)
        dh = np.max(self.zaxis.value) - np.min(self.zaxis.value)

        fig = bkp.figure(plot_width=1000, plot_height=1000,
                         title=f'{self.inputs.options.species} {rname}',
                         x_axis_label=f'Distance ({ustr})',
                         y_axis_label=f'Distance ({ustr})',
                         x_range=[np.min(self.xaxis.value),
                                  np.max(self.xaxis.value)],
                         y_range=[np.min(self.zaxis.value),
                                  np.max(self.zaxis.value)],
                         tooltips=tooltips)

        fig.image(image=[self.image], x=x0, y=y0, dw=dw, dh=dh,
                  color_mapper=color_mapper)
        xc = np.cos(np.linspace(0, 2*np.pi, 1000))
        yc = np.sin(np.linspace(0, 2*np.pi, 1000))
        fig.patch(xc, yc, fill_color='yellow')
        
        bkp.output_file(savefile)
        if show:
            bkp.show(fig)
        else:
            bkp.save(fig)
            
        assert 0

        # from astropy.visualization import (PercentileInterval, LogStretch,
        #                                    ImageNormalize)
        #
        # # Determine limits if none given
        # if limits is None:
        #     interval = PercentileInterval(95)
        #     self.blimits = interval.get_limits(self.image[self.image > 0])
        # elif len(limits) == 2:
        #     self.blimits = limits
        # else:
        #     assert 0, 'Problem with the display limits'

#        norm = ImageNormalize(self.image, stretch=LogStretch(),
#                              vmin=self.blimits[0], vmax=self.blimits[1])

        # # Make the colorbar
        # if self.quantity == 'column':
        #     clabel = f'$N_{{ {self.inputs.options.species} }}\ cm^{{-2}}$'
        # else:
        #     clabel = f'$I_{{ {self.inputs.options.species} }} R$'
        # cbar = fig.colorbar(im, shrink=0.7, label=clabel)

        # Put Planet's disk in the middle
        # xc, yc = (np.cos(np.linspace(0, 2*np.pi, 1000)),
        #           np.sin(np.linspace(0, 2*np.pi, 1000)))
        # ax.fill(xc, yc, 'y')

        return fig
    
    def image_rotation(image):
        slong = image.subobslongitude
        slat = image.subobslatitude
        
        pSun = np.array([0., -1., 0.])
        pObs = np.array([np.sin(slong)*np.cos(slat),
                         -np.cos(slong)*np.cos(slat),
                         np.sin(slat)])
        if np.array_equal(pSun, pObs):
            M = npmat.identity(3)
        else:
            costh = np.dot(pSun, pObs)/np.linalg.norm(pSun)/np.linalg.norm(pObs)
            theta = np.arccos(np.clip(costh, -1, 1))
            axis = np.cross(pSun, pObs)
            M = rotation_matrix(theta, axis)
        
        #M = np.transpose(M)
        return M





#!/usr/bin/env python
# Licensed under a 3-clause BSD style license - see LICENSE.rst
import glob
import os
import subprocess
import sys
from setuptools import setup, find_packages, Extension, Command
from setuptools.command.test import test as TestCommand

try:
    from configparser import ConfigParser
except ImportError:
    from ConfigParser import ConfigParser

# Get some values from the setup.cfg
conf = ConfigParser()
conf.read(['setup.cfg'])
metadata = dict(conf.items('metadata'))

PACKAGENAME = metadata.get('package_name', 'nexoclom')
DESCRIPTION = metadata.get('description', '')
AUTHOR = metadata.get('author', 'Matthew Burger')
AUTHOR_EMAIL = metadata.get('author_email', 'mburger@stsci.edu')
LICENSE = metadata.get('license', 'unknown')
URL = metadata.get('url', 'unknown')
__minimum_python_version__ = metadata.get("minimum_python_version", "3.7")

# if os.path.exists('relic'):
#     sys.path.insert(1, 'relic')
#     import relic.release
# else:
#     try:
#         # Note: This is the way this will work
#         import relic.release
#     except ImportError:
#         try:
#             subprocess.check_call(['git', 'clone',
#                                    'https://github.com/jhunkeler/relic.git'])
#             sys.path.insert(1, 'relic')
#             import relic.release
#         except subprocess.CalledProcessError as e:
#             print(e)
#             exit(1)

#version = relic.release.get_info()
#relic.release.write_template(version, PACKAGENAME)
for line in open(f'{PACKAGENAME}/__init__.py', 'r').readlines():
    if 'version' in line:
        version = line.split('=')[1].strip().replace("'", "").replace('"', '')


# allows you to build sphinx docs from the pacakge
# main directory with python setup.py build_sphinx
try:
    from sphinx.cmd.build import build_main
    from sphinx.setup_command import BuildDoc

    class BuildSphinx(BuildDoc):
        """Build Sphinx documentation after compiling C source files"""

        description = 'Build Sphinx documentation'

        def initialize_options(self):
            BuildDoc.initialize_options(self)

        def finalize_options(self):
            BuildDoc.finalize_options(self)

        def run(self):
            build_cmd = self.reinitialize_command('build_ext')
            build_cmd.inplace = 1
            self.run_command('build_ext')
            build_main(['-b', 'html', './docs', './docs/_build/html'])

except ImportError:
    class BuildSphinx(Command):
        user_options = []

        def initialize_options(self):
            pass

        def finalize_options(self):
            pass

        def run(self):
            print('!\n! Sphinx is not installed!\n!', file=sys.stderr)
            exit(1)


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = ['nexoclom/tests']
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)


readme_glob = 'README*'
if os.path.exists('LONG_DESCRIPTION.rst'):
    with open('LONG_DESCRIPTION.rst') as f:
        LONG_DESCRIPTION = f.read()
elif len(glob.glob(readme_glob)) > 0:
    with open(glob.glob(readme_glob)[0]) as f:
        LONG_DESCRIPTION = f.read()
else:
    # Get the long description from the package's docstring
    __import__(PACKAGENAME)
    package = sys.modules[PACKAGENAME]
    LONG_DESCRIPTION = package.__doc__

setup(name=PACKAGENAME,
      version=version,
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      license=LICENSE,
      url=URL,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      install_requires= [s.strip() for s in
                 metadata.get('install_requires', 'astropy').split(',')],
      python_requires='>={}'.format(__minimum_python_version__),
      tests_require=['pytest'],
      packages=find_packages(),
      package_data={PACKAGENAME: ['data/*']},
      include_package_data=True,
      cmdclass={'test': PyTest,
                'build_sphinx': BuildSphinx},
)



"""Tests for source_distribution.py
Functions to test:
    gaussiandist (Not tested - this isn't used)
    sputdist
    MaxwellianDist
    xyz_from_lonlat
    surface_distribution
    surface_spot
    idlversion
    speed_distribution
    angular_distribution

Distributions to be compared with results from IDL version of code"""
import numpy as np
import os.path
from scipy.io import readsav
import astropy.units as u
import bokeh.plotting as plt
try:
    from ..source_distribution import *
except ValueError:
    from nexoclom.source_distribution import *

TOL = 1e-5


def test_maxwellian_dist():
    idl = readsav(os.path.join(os.path.dirname(__file__), 'data',
                               'maxdist.sav'))
    vv =  idl['vv']*u.km/u.s
    temperature = [t*u.K for t in idl['temp']]
    atom = [str(a)[2:-1] for a in idl['atom']]
    f_idl = idl['max_dist']

    for i,t in enumerate(temperature):
        for j,a in enumerate(atom):
            f_idl_ = np.array(f_idl[j,i,:], dtype=float)
            f_idl_ /= np.max(f_idl_)
            f_py = MaxwellianDist(vv, t, a)

            print(np.max(np.abs(f_idl_-f_py)))
            assert np.all(np.abs(f_idl_-f_py) < TOL), (
                f'atom={a}, temperature={t}')

def test_sputdist():
    idl = readsav(os.path.join(os.path.dirname(__file__), 'data',
                               'sputdist.sav'))
    vv =  idl['vv']*u.km/u.s
    atom = [str(a)[2:-1] for a in idl['atom']]
    alpha = [a for a in idl['alpha']]
    bet = [b for b in idl['bet']]
    U = [u_*u.eV for u_ in idl['U']]
    f_idl = idl['sput_dist']

    for i,al in enumerate(alpha):
        for j,b in enumerate(bet):
            for k,u_ in enumerate(U):
                for l,a in enumerate(atom):
                    f_idl_ = np.array(f_idl[l,k,j,i,:], dtype=float)
                    f_idl_ /= np.max(f_idl_)
                    f_py = sputdist(vv, u_, al, b, a)

            print(np.max(np.abs(f_idl_-f_py)))
            assert np.all(np.abs(f_idl_-f_py) < TOL), (
                f'atom={a}, alpha={al}, beta={b}, U={u_}')


if __name__ == '__main__':
    test_maxwellian_dist()
    test_sputdist()



import matplotlib
#matplotlib.use("Agg")
import matplotlib.pyplot as plt
import matplotlib.animation as animation

import numpy as np
from nexoclom import Input, Output
from MESSENGERuvvs import MESSENGERdata
import pickle
from mathMB import minmaxmean
import time

def test_compareversions():
    # old version
    old = pickle.load(open('data/oldmodel.pkl', 'rb'))
    old0 = pickle.load(open('data/oldmodel0.pkl', 'rb'))

    # new version
    npack = 1e5
    orbit = 36
    overwrite = True

    inputsfile = 'inputfiles/Na.Gaussian.3_1.input'
#    inputsfile = 'inputfiles/Na.v3.radial.input'
    inputs = Input(inputsfile)
    inputs.run(npack, overwrite=overwrite)

    # imformat = 'MercuryEmission.format'
    # image = inputs.produce_image(imformat)
    # image.display(savefile=inputsfile.replace('.input', '_radiance.png'),
    #               show=False)
    #
    # colformat = 'MercuryColumn.format'
    # column = inputs.produce_image(colformat)
    # column.display(savefile=inputsfile.replace('.input', '_column.png'),
    #                show=False)
    
    savefile, _, _ = inputs.search()
    output = Output.restore(savefile[0])
    X = output.X
    X0 = output.X0
    
    # xbin = np.linspace(-1, 1, 25)
    # h, _ = np.histogram(X0.x, bins=xbin)
    # h0, _ = np.histogram(old0.x, bins=xbin)
    # plt.plot(xbin[0:-1], h)
    # plt.plot(xbin[0:-1], h0)
    # plt.show()
    #
    # h, _ = np.histogram(X0.y, bins=xbin)
    # h0, _ = np.histogram(old0.y, bins=xbin)
    # plt.plot(xbin[0:-1], h)
    # plt.plot(xbin[0:-1], h0)
    # plt.show()
    #
    # h, _ = np.histogram(X0.z, bins=xbin)
    # h0, _ = np.histogram(old0.z, bins=xbin)
    # plt.plot(xbin[0:-1], h)
    # plt.plot(xbin[0:-1], h0)
    # plt.show()
    #
    # xbin = np.linspace(-6, 6, 100)
    # h, _ = np.histogram(X0.vx*2440, bins=xbin)
    # h0, _ = np.histogram(old0.vx*2440, bins=xbin)
    # plt.plot(xbin[0:-1], h)
    # plt.plot(xbin[0:-1], h0)
    # plt.show()
    #
    # h, _ = np.histogram(X0.vy*2440, bins=xbin)
    # h0, _ = np.histogram(old0.vy*2440, bins=xbin)
    # plt.plot(xbin[0:-1], h)
    # plt.plot(xbin[0:-1], h0)
    # plt.show()
    #
    # h, _ = np.histogram(X0.vz*2440, bins=xbin)
    # h0, _ = np.histogram(old0.vz*2440, bins=xbin)
    # plt.plot(xbin[0:-1], h)
    # plt.plot(xbin[0:-1], h0)
    # plt.show()
    # assert 0

    # Set up formatting for the movie files
    
    animate(X, 'newmodel.mp4')
#    animate(old, 'oldmodel.mp4')



def animate(data, filename):
    Writer = animation.writers['ffmpeg']
    writer = Writer(fps=15, metadata=dict(artist='Me'), bitrate=1800)
    
    def xyscatter(t=0):
        while t <= len(times):
            t += 1
            x = data.x[data.time == times[t-1]].values
            y = data.y[data.time == times[t-1]].values
            # f = data.frac[data.time == times[t-1]].values
            yield x, y
            
    def init():
        ax.set_xlim(-10,10)
        ax.set_ylim(-10,10)
        ax.set_aspect('equal')
        ax.plot(np.cos(np.linspace(0, 2*np.pi, 361)),
                np.sin(np.linspace(0, 2*np.pi, 361)))
                
        line.set_data(xdata, ydata)
        
        return line,
        
    def run(data):
        xdata, ydata = data
        line.set_data(xdata, ydata)

    times = data.time.unique()
    t = times[0]
    xdata, ydata = data.x[data.time == t], data.y[data.time == t]

    fig, ax = plt.subplots(1, 1)
    line, = ax.plot(xdata, ydata, marker='.', linestyle=' ')
    
    ani = animation.FuncAnimation(fig, run, xyscatter, interval=10,
                                  init_func=init)
    ani.save(filename, writer=writer)
    plt.close()


if __name__ == '__main__':
    test_compareversions()



from .Input import Input
from .Output import Output
from .LOSResult import LOSResult
from .produce_image import ModelImage
from .configure_model import configure_model
from .LossInfo import LossInfo
from .LOSResult import LOSResult
from .IDLout import IDLout
from .database_connect import database_connect


name = 'nexoclom'
__author__ = 'Matthew Burger'
__email__ = 'mburger@stsci.edu'
__version__ = '2.1.7'

configure_model()
